#! /usr/bin/python3.2
# (home + travail + magi)

# ============================================================================== 
#						
#		  		UTILITIES FUNCTIONS FOR PYTON SCRIPTS RELATED TO
#				LEAR, TILDE AND NFOIL
#
# ==============================================================================


# =================================== IMPORTS ==================================

import os
import stats_learning

# ================================= FUNCTIONS ==================================

def printList(message, uneListe) :
	print(message)
	for element in uneListe :
		print(element)

								# get results of the learning with Nfoil
def getNfoilResults(fileResults):
							
		fResults = open(fileResults, "r")
		lines = fResults.readlines()
		fResults.close()

		nbLines = len(lines)
		
								# the results are in the four last lines

		ligneTruePos = lines[nbLines-4]
		elementsTruePos = ligneTruePos.split()
		truePos = elementsTruePos[3]

		ligneFalseNeg = lines[nbLines-3]
		elementsFalseNeg = ligneFalseNeg.split()
		falseNeg = elementsFalseNeg[3]

		ligneTrueNeg = lines[nbLines-2]
		elementsTrueNeg = ligneTrueNeg.split()
		trueNeg = elementsTrueNeg[3]

		ligneFalsePos = lines[nbLines-1]
		elementsFalsePos = ligneFalsePos.split()
		falsePos = elementsFalsePos[3]
		
		return (truePos, falseNeg, trueNeg, falsePos)	
			

								# get results of the learning with Tilde
def getTildeResults(file_ambiguity):
							
		fileResults = os.path.join("./tilde", file_ambiguity + "_tilde.summary.uL")
		fTildeResults = open(fileResults, "r")
		lines = fTildeResults.readlines()
		fTildeResults.close()

		nbLines = len(lines)
		
								# get true pos and false  
		lignePos = lines[nbLines-9]
		elementsPos = lignePos.split()
		truePos = elementsPos[2]
		falseNeg = elementsPos[3]

								# get true neg and false pos
		ligneNeg = lines[nbLines-8]
		elementsNeg = ligneNeg.split()
		falsePos = elementsNeg[2]
		trueNeg  = elementsNeg[3]
		
		return (truePos, falseNeg, trueNeg, falsePos)	


		# compute the final results for the current probability
def computeAverageResults(fileExcelProba) :
			
	f = open(fileExcelProba + ".csv", "r")
	lines = f.readlines()
	f.close()
	
	accuraciesPlus    = []
	accuraciesMoins   = []
	accuracies        = []
	accuraciesMean    = []
	
						# each line is truePos;falseNeg;TrueNeg;falsePos
						# for the current Fold
	for line in lines :
		
		elements  = line.split(';')
		
		accuracyPlus  =  float(elements[0])
		accuracyMoins =  float(elements[1])
		accuracy      =  float(elements[2])
		accuracyMean  =  float(elements[3])
		
		accuraciesPlus.append(accuracyPlus)
		accuraciesMoins.append(accuracyMoins)
		accuracies.append(accuracy)
		accuraciesMean.append(accuracyMean)
			
	f = open(fileExcelProba + "_final.csv", "w")
	f.write(
		    str(stats_learning.moyenne(accuraciesPlus))  + ";" +
		    str(stats_learning.moyenne(accuraciesMoins)) + ";" +
		    str(stats_learning.moyenne(accuracies))      + ";" +
		    str(stats_learning.moyenne(accuraciesMean))
		    )
	f.close()	
		
		
		# return results of nbFolds of a cross validation
def getCrossValidationResults(fileExcel):
	
	f = open(fileExcel, "r")
	lines = f.readlines()
	f.close()
		
	accuraciesPlus    = []
	accuraciesMoins   = []
	accuracies        = []
	accuraciesMean    = []
	
							# each line is truePos;falseNeg;TrueNeg;falsePos
							# for the current Fold
	for line in lines :
		
		elements  = line.split(';')
		
		truePos  =  int(elements[0])
		falseNeg =  int(elements[1])
		trueNeg  =  int(elements[2])
		falsePos =  int(elements[3])
		
		accuracyPlus  = stats_learning.accuracy(truePos, falseNeg)
		accuracyMoins = stats_learning.accuracy(trueNeg, falsePos)
		accuracy      = stats_learning.accuracy(truePos + trueNeg, falsePos + falseNeg)
		accuracyMean  = stats_learning.moyenne([accuracyPlus, accuracyMoins])
		
		accuraciesPlus.append(accuracyPlus)
		accuraciesMoins.append(accuracyMoins)
		accuracies.append(accuracy)
		accuraciesMean.append(accuracyMean)
		
	return (
		     stats_learning.moyenne(accuraciesPlus), 
		     stats_learning.moyenne(accuraciesMoins),
		     stats_learning.moyenne(accuracies),
		     stats_learning.moyenne(accuraciesMean)
		  )

		# add the means of accuracyPlus, accuracyMoins, accuracy, accuracyMean
		# of the folds of the current crossValidation
def addCrossValidationResults(fileExcel, results) :
	
	(accuracyPlusAVG, accuracyMoinsAVG, accuracyAVG, accuracyMeanAVG) = results
	
	f = open(fileExcel, "a")
	f.write(str(accuracyPlusAVG) + ";" + str(accuracyMoinsAVG) + ";" +
		    str(accuracyAVG)     + ";" + str(accuracyMeanAVG)  + "\n")		
	f.close()
		
# ================================== END =======================================

