#!/usr/bin/python3.2
# (home + travail + magi)

# ==============================================================================
#                       CREATE N CONCURRENT COMPLETIONS
#                       Lear ambiguous intensional examples (.int) 
#						are completed using certain (non ambiguous) data					   
#
#	script     = sys.argv[0]       # nom du script
#	fileEx     = sys.argv[1]       # file containing train intentionsal complete examples (.int)
#	fileBkg	 = sys.argv[2]		 # file containing background knowledge (.int_bkg)
#	nbCross    = int(sys.argv[3])  # cross validation identifier
#	nbFold     = int(sys.argv[4])  # number of folds of the cross validation 
#	ambiguity  = sys.argv[5]       # level of ambiguity (percentage)

# =================================== IMPORTS ==================================

import sys
import os
import multiprocessing as mp
import subprocess

# =============================== VARIABLES GLOBALES ===========================

learDirectory = os.getenv('LEAR_DIR')
LEAR = os.path.join(learDirectory, "topLear3.pl")

# =================================== FONCTIONS ================================
				
	# make a nbFolds cross validation
def crossValidation(fileEx, file_int_bkg, cross, nbFold, ambiguity) :

				# we are in the proba directory
	suffixCross = "_n" + str(cross)
	currentDirectory = os.getcwd()
	baseName = os.path.basename(currentDirectory)
	baseCrossDirectory = baseName + suffixCross
	crossDirectory = os.path.join(currentDirectory, baseCrossDirectory)
	os.chdir(crossDirectory)	
	
					# we are in the cross directory
	processName = mp.current_process().name
	pOut = open(processName + ".out", "w")
	pErr = open(processName + ".err", "w")
         						     		               	
	for fold in range(1, nbFold+1) :
			
		print("fold : ", fold),
		foldBase = baseCrossDirectory + "_fold" + str(fold)
							
					# go to the directory of the current fold
		foldDirectory = os.path.join(crossDirectory, foldBase) 
		os.chdir(foldDirectory)

					# build the name of the ambiguity file name
		file_ambiguity_int  = foldBase + "_train_ambiguous"
      
					# file intensional ambiguous examples are completed 
		file_ambiguity_int_completed = file_ambiguity_int + "_close_world"
		args = [LEAR, "-X", file_ambiguity_int, "-B", file_int_bkg, 
			   "-o", file_ambiguity_int_completed]	     
		subprocess.call(args, stdout = pOut, stderr = pErr)

			
		os.chdir("..")	# return to the cross directory		
		
	os.chdir("..")		# return to the proba directory
		
	pOut.close()
	pErr.close()


# ===================================== MAIN ===================================

if __name__ == '__main__' :

	script     = sys.argv[0]       # nom du script
	fileEx     = sys.argv[1]       # file containing train intentionsal complete examples (.int)
	fileBkg	   = sys.argv[2]	 # file containing background knowledge (.int_bkg)
	nbCross    = int(sys.argv[3])  # cross validation identifier
	nbFold     = int(sys.argv[4])  # number of folds of the cross validation 
	ambiguity  = sys.argv[5]       # level of ambiguity (percentage)

	print("begin of the script ", script)
	print("current directory = " + os.getcwd())

	suffixProba = "_p" + ambiguity
	directoryProba = fileEx    + suffixProba
	os.chdir(directoryProba)	

							# we are in the proba directory
					
	processes = []

							 # creation des process crossValidation
	for cross in range(0, nbCross) :	
		
		processes.append(
			             mp.Process(
						               target = crossValidation, 
						               name = "close_world_completion_n" + str(cross), 
						               args = (
							                   fileEx, fileBkg,
                                               cross, nbFold, ambiguity
							                  )
                                )
                          )   						

							# running the nbCross cross validations	  
	for p in processes : 
		p.start()      

							# waiting for the cross validations terminate
	for p in processes : 	
	 	p.join()       

print("end of the script ", script)

# ====================================== END ===================================
