#!/usr/bin/python3.2
# (home + travail)

#!/usr/local/python32/bin/python3.2
# (magi)

# ==============================================================================
#                           CREATE N CONCURRENT K CROSS-VALIDATIONS
#                           (here only the building of extensional examples
#                            and the learning are done)						   
#
#	script     = sys.argv[0]       	# nom du script
#	fileEx     = sys.argv[1]       	# file containing train intentionsal complete examples (.int)
#	fileBkg    = sys.argv[2]       	# file containing background knowledge
#	fileExcel  = sys.argv[3]       	# prefix of files that will contain results in excel format
#	nbCross    = int(sys.argv[4])  	# cross validation identifier
#	nbFold     = int(sys.argv[5])  	# number of folds of the cross validation
#	ambiguity  = sys.argv[6]       	# level of ambiguity
#    nbErrors   = sys.argv[7]      	# number of accepted errors on negative examples
#    nbTries    = sys.argv[8]      	# number of tries to enhance an hypothesis
#                                   	# having less than nbErrors
#	compWeight = sys.argv[9]	          # weight of the only compatible examples

# =================================== IMPORTS ==================================

import sys
import os
import multiprocessing as mp
import subprocess
import lear_script_utilities as utilities

# =============================== VARIABLES GLOBALES ===========================

learDirectory = os.getenv('LEAR_DIR')
LEAR = os.path.join(learDirectory, "topLear3.pl")
lockFileProba = mp.Lock()     	

# =================================== FONCTIONS ================================
	
					# print system arguments
def printSystemArguments():
	for argument in sys.argv:
		print(argument)


					# make a nbFold cross validation
def crossValidation(fileEx, fileBkg, fileExcel, fileExcelProba,
       			cross, nbFold, ambiguity,
                    nbErrors, nbTries, compWeight) :

	processName = mp.current_process().name
	pOut = open(processName + ".out", "w")
	pErr = open(processName + ".err", "w")

	suffixCross = "_n" + str(cross)
	fileExcel = fileExcel + suffixCross

					# we are in the proba directory

	currentDirectory = os.getcwd() ;
	baseName = os.path.basename(currentDirectory)
	baseCrossDirectory = baseName + suffixCross

					# go to the cross validation directory
	crossDirectory = os.path.join(currentDirectory, baseCrossDirectory)
	os.chdir(crossDirectory)	

					# for each fold	     			               	         						     		               	
	for fold in range(1, nbFold+1) :
			
		print("fold : ", fold),
		foldBase = baseCrossDirectory + "_fold" + str(fold)
							
					# go to the current fold directory
		foldDirectory = os.path.join(crossDirectory, foldBase) 
		os.chdir(foldDirectory)
	
					# copy file containing the background knowledge
#		os.system("cp ../../../" + fileBkg + ".int_bkg .")   	
		
		fileDirectory = os.path.basename(foldDirectory)
		
					# build file names
		fileTrainComplete = fileDirectory + "_train" 
		fileTest          = fileDirectory + "_test"		
		fileAmbiguity= fileDirectory+ "_train_ambiguous_close_world" 
    		
					# build extensional examples from the ambiguity file
		args = [LEAR, "-e", fileAmbiguity, "-B", fileBkg]	
		subprocess.call(args, stdout = pOut, stderr = pErr)
       											              		
					# learn from ambiguity file and test on test file
		args = [LEAR, "-f", fileAmbiguity, "-u", fileTest, "-g",
		   		"../" + fileExcel, "-w", "5", 
		   		"-v", "3", "-p", nbErrors, "-k", nbTries, "-a", "-j", compWeight]
		subprocess.call(args, stdout = pOut, stderr = pErr)

					# return to the cross directory	
		os.chdir("..")	

					# get the average of the results of the nbFold folds
	crossValidationResults = utilities.getCrossValidationResults(fileExcel + ".csv")

					# return to the proba directory
	os.chdir("..")		
		
					# save results
	lockFileProba.acquire()
	utilities.addCrossValidationResults(fileExcelProba, crossValidationResults)
	lockFileProba.release()

	pOut.close()
	pErr.close()

# ===================================== MAIN ===================================

if __name__ == '__main__' :

	script     = sys.argv[0]       # script name
	fileEx     = sys.argv[1]       # file containing train intentionsal complete examples (.int)
	fileBkg    = sys.argv[2]       # file containing background knowledge
	fileExcel  = sys.argv[3]       # prefix of files that will contain results in excel format
	nbCross    = int(sys.argv[4])  # cross validation identifier
	nbFold     = int(sys.argv[5])  # number of folds of the cross validation
	ambiguity  = sys.argv[6]       # level of ambiguity    
	nbErrors   = sys.argv[7]       # number of accepted errors on negative examples
	nbTries    = sys.argv[8]       # number of tries to enhance an hypothesis
                                    # having less than nbErrors
	compWeight = sys.argv[9]	      # weight of the only compatible examples	

	print("begin of the script ", script)

	suffixProba = "_p" + ambiguity
	fileExcelProba = fileExcel + suffixProba + ".csv"
	directoryProba = fileEx    + suffixProba

							# go to the proba directory
	os.chdir(directoryProba)	
					
	processes = []

							# creation des process crossValidation
	for cross in range(0, nbCross) :	
		processes.append(
			             mp.Process(
						         target = crossValidation, 
						         name = "ext_lear_n" + str(cross), 
						         args = (
							             fileEx, fileBkg, 
                                                fileExcel, fileExcelProba,
                                                cross, nbFold, ambiguity,
                                                nbErrors, nbTries, compWeight
							            )
                                     )
                          )   						

							# running the cross validations	  
	for p in processes : 
		p.start()      
							# waiting for the cross validations terminate
	for p in processes : 	
	 	p.join()       

	print("end of the script ", script)

# ====================================== END ===================================
