#!/usr/bin/python3.2
# (home + travail)

#!/usr/local/python32/bin/python3.2
# (magi)

# ==============================================================================
#                           CREATE N CONCURRENT K CROSS-VALIDATIONS	
#                           ON FILES ALREADY TREATED
#                           THAT WE RETREATED WITH OTHER BACKGROUND KNOWLEDGE					   
#
#	script     = sys.argv[0]       # nom du script
#	fileEx     = sys.argv[1]       # file containing train intentionsal complete examples (.int)
#	fileBkg    = sys.argv[2]       # file containing background knowledge
#	fileExcel  = sys.argv[3]       # prefix of files that will contain results in excel format
#	nbCross    = int(sys.argv[4])  # cross validation identifier
#	nbFold     = int(sys.argv[5])  # number of folds of the cross validation
#	factor     = sys.argv[6]       # create factor ambiguous examples from each complete example    
#	ambiguity  = sys.argv[7]       # level of ambiguity (percentage)
#    nbErrors   = sys.argv[8]       # number of accepted errors on negative examples
#    nbTries    = sys.argv[9]       # number of tries to enhance an hypothesis
                                    # having less than nbErrors

# =================================== IMPORTS ==================================

import sys
import os
import multiprocessing as mp
import subprocess
import lear_script_utilities as utilities

# =============================== VARIABLES GLOBALES ===========================

learDirectory = os.getenv('LEAR_DIR')
LEAR = os.path.join(learDirectory, "topLear3.pl")
print("lear directory =" + learDirectory)
print("LEAR=" + LEAR)
lockFileProba = mp.Lock()     	

# =================================== FONCTIONS ================================
	
	# make a nbFolds cross validation
def crossValidation(fileEx, fileBkg, fileExcel, fileExcelProba,
       			cross, nbFold, factor, ambiguity,
                    nbErrors, nbTries) :

	processName = mp.current_process().name
	pOut = open(processName + ".out", "w")
	pErr = open(processName + ".err", "w")

	suffixCross = "_n" + str(cross)
	fileExcel = fileExcel + suffixCross

					# we are in the proba directoy
	currentDirectory = os.getcwd() ;
	baseName = os.path.basename(currentDirectory)
	baseCrossDirectory = baseName + suffixCross
	crossDirectory = os.path.join(currentDirectory, baseCrossDirectory)
	os.chdir(crossDirectory)	
	     			  
					# we are in the cross directory
       	         						     		               	
	for fold in range(1, nbFold+1) :
			
		print("fold : ", fold),
		foldBase = baseCrossDirectory + "_fold" + str(fold)
							
					# go to the directory of the current fold
		foldDirectory = os.path.join(crossDirectory, foldBase) 
		os.chdir(foldDirectory)
		
					# copy file containing the new background knowlege
		os.system("cp ../../../" + fileBkg + ".int_bkg .")    
		
		fileDirectory = os.path.basename(foldDirectory)
		
		fileTrainComplete = fileDirectory + "_train_complete"  # complete train examples
		fileTest          = fileDirectory + "_test"		
		 
        		
    					# build name of the ambiguity file
		fileAmbiguity= fileDirectory + "_train" 
    		      		
					# build extensional examples from the ambiguity file
		args = [LEAR, "-e", fileAmbiguity, "-B", fileBkg]	
		subprocess.call(args, stdout = pOut, stderr = pErr)
       											              		
					# learn from ambiguity file and test on test file
		args = [LEAR, "-f", fileAmbiguity, "-u", fileTest, "-g",
		   		"../" + fileExcel, "-w", "5", 
		   		"-v", "3", "-p", nbErrors, "-k", nbTries, "-a", "-j", "0.75"]
		subprocess.call(args, stdout = pOut, stderr = pErr)

		os.chdir("..")	# return to the cross directory	

					# get the average of the results of the nbFold folds
	crossValidationResults = utilities.getCrossValidationResults(fileExcel + ".csv")

	os.chdir("..")		# return to the proba directory
		

					# save results for this n
	lockFileProba.acquire()
	utilities.addCrossValidationResults(fileExcelProba, crossValidationResults)
	lockFileProba.release()

	pOut.close()
	pErr.close()

# ===================================== MAIN ===================================

if __name__ == '__main__' :

	script     = sys.argv[0]       # nom du script
	fileEx     = sys.argv[1]       # file containing train intentionsal complete examples (.int)
	fileBkg    = sys.argv[2]       # file containing background knowledge
	fileExcel  = sys.argv[3]       # prefix of files that will contain results in excel format
	nbCross    = int(sys.argv[4])  # cross validation identifier
	nbFold     = int(sys.argv[5])  # number of folds of the cross validation
	factor     = sys.argv[6]       # create factor ambiguous examples from each complete example    
	ambiguity  = sys.argv[7]       # level of ambiguity (percentage)
	nbErrors   = sys.argv[8]       # number of accepted errors on negative examples
	nbTries    = sys.argv[9]       # number of tries to enhance an hypothesis
                                    # having less than nbErrors

	print("begin of the script ", script)

	print("nbErrors = ", nbErrors)
	print("nbTries = ", nbTries)

	suffixProba = "_p" + ambiguity
	fileExcelProba = fileExcel + suffixProba + ".csv"
	directoryProba = fileEx    + suffixProba

	os.chdir(directoryProba)	
					
	processes = []

							# creation des process crossValidation
	for cross in range(0, nbCross) :	
		processes.append(
			             mp.Process(
						         target = crossValidation, 
						         name = "crossValidation_n" + str(cross), 
						         args = (
							             fileEx, fileBkg, 
                                                fileExcel, fileExcelProba,
                                                cross, nbFold, factor, ambiguity,
                                                nbErrors, nbTries
							            )
                                     )
                          )   						

							# running the cross validations	  
	for p in processes : 
		p.start()      

							# waiting for the cross validations terminate
	for p in processes : 	
	 	p.join()       

	print("end of the script ", script)

# ====================================== END ===================================
