#!/usr/bin/python3.2
# (home + travail + magi)

# ==============================================================================
#                           CREATE N CONCURRENT K CROSS-VALIDATIONS						   
#
#	script     = sys.argv[0]       	# nom du script
#	fileEx     = sys.argv[1]       	# file containing train intentionsal complete examples (.int)
#	fileBkg    = sys.argv[2]       	# file containing background knowledge
#	fileExcel  = sys.argv[3]       	# prefix of files that will contain results in excel format
#	nbCross    = int(sys.argv[4])  	# cross validation identifier
#	nbFold     = int(sys.argv[5])  	# number of folds of the cross validation
#	factor     = sys.argv[6]       	# create factor ambiguous examples from each complete example   
#	ambiguity  = sys.argv[7]        	# level of ambiguity (percentage) 
#    nbErrors   = sys.argv[8]      	# number of accepted errors on negative examples
#    nbTries    = sys.argv[9]      	# number of tries to enhance an hypothesis
#                                   	# having less than nbErrors
#	compWeight = sys.argv[10]	     # weight of the only compatible examples
#	doubleRep  = sys.argv[11]		# double representation (double_representation) 

# =================================== IMPORTS ==================================

import sys
import os
import multiprocessing as mp
import subprocess
import lear_script_utilities as utilities

# =============================== VARIABLES GLOBALES ===========================

learDirectory = os.getenv('LEAR_DIR')
LEAR = os.path.join(learDirectory, "topLear3.pl")
lockFileProba = mp.Lock()     	

# =================================== FONCTIONS ================================
	
					# make a nbFold cross validation
def crossValidation(fileEx, fileBkg, fileExcel, fileExcelProba,
       			cross, nbFold, factor, ambiguity,
                    nbErrors, nbTries, compWeight, doubleRep) :

	processName = mp.current_process().name

	suffixCross = "_n" + str(cross)
	fileExcelCross = fileExcel + suffixCross

					# we are in the proba directory

	currentDirectory = os.getcwd()
	baseName = os.path.basename(currentDirectory)
	baseCrossDirectory = baseName + suffixCross

					# create a new cross validation directory
	crossDirectory = os.path.join(currentDirectory, baseCrossDirectory)
	os.mkdir(crossDirectory)
	os.chdir(crossDirectory)	

					# we are in the new cross validation directory
	pOut = open(processName + ".out", "w")
	pErr = open(processName + ".err", "w")	     			     
          	         						     		               	
	for fold in range(1, nbFold+1) :
			
		print("fold : ", fold),
		foldBase = baseCrossDirectory + "_fold" + str(fold)
							
					# create a directory for the current fold
		foldDirectory = os.path.join(crossDirectory, foldBase) 
		os.mkdir(foldDirectory)
		os.chdir(foldDirectory)
					
					# we are in the fold directory					
					
					# copy file containing the examples
		os.system("cp ../../../" + fileEx   + ".int . ")	
	
					# copy file containing the background 
		os.system("cp ../../../" + fileBkg + ".int_bkg .")   	
		
		fileDirectory = os.path.basename(foldDirectory)
		
					# (complete) train examples
		fileTrainComplete = fileDirectory + "_train" 
		fileTest          = fileDirectory + "_test"		

					# create (complete) train_fold and test_fold files.
		print("creating train and test files...")
		program = os.path.join(learDirectory, "create_train_test_files.pl")
		subprocess.call([program, fileEx, fileTrainComplete, 
					  fileTest, str(fold), str(nbFold)],
					  stdout = pOut, stderr = pErr) 	 

    					# build extensional examples from the test file
		print("building extensional test examples...")   
		args = [LEAR, "-e", fileTest, "-B", fileBkg]	
		subprocess.call(args, stdout = pOut, stderr = pErr)
    		
    					# create the ambiguous examples
		print("building train ambiguous examples...")
		fileAmbiguity= fileDirectory+ "_train_ambiguous" 
		program = os.path.join(learDirectory, "ambiguity_generator.pl")
		args = [program, fileTrainComplete, fileAmbiguity, 
			   ambiguity, factor, doubleRep]	

		subprocess.call(args, stdout = pOut, stderr = pErr) 
      		
					# build extensional examples from the ambiguity file
					# this create four files :
					# - fileAmbiguity.pl containing extensional examples
					#   (some intentional examples have been not considered).
					# - fileAmbiguity.bkg containing background knowledge.
					# - fileAmbiguity_remaining.int containing intentional
					#   examples that have been traslated in extensional ones.
					#   these examples will be used by Tilde and Nfoil.
 
					# compute train extensional ambiguous examples
		print("building extensional train ambiguous examples...")
		args = [LEAR, "-e", fileAmbiguity, "-B", fileBkg]	
		subprocess.call(args, stdout = pOut, stderr = pErr)
       											              		
					# learn from ambiguity file and test on test file
		print("learning...")
		args = [LEAR, "-f", fileAmbiguity, "-u", fileTest, "-g",
		   		"../" + fileExcelCross, "-w", "5", 
		   		"-v", "1", "-p", nbErrors, "-k", nbTries, "-a", "-j", compWeight]
		subprocess.call(args, stdout = pOut, stderr = pErr)

					# return to the cross directory	
		os.chdir("..")	

					# get the average of the results of the nbFold folds
	crossValidationResults = utilities.getCrossValidationResults(fileExcelCross + ".csv")

					# return to the proba directory
	os.chdir("..")		
		
					# save results
	lockFileProba.acquire()
	utilities.addCrossValidationResults(fileExcelProba, crossValidationResults)
	lockFileProba.release()

	pOut.close()
	pErr.close()

# ===================================== MAIN ===================================

if __name__ == '__main__' :

	script     = sys.argv[0]       # script name
	fileEx     = sys.argv[1]       # file containing train intentionsal complete examples (.int)
	fileBkg    = sys.argv[2]       # file containing background knowledge
	fileExcel  = sys.argv[3]       # prefix of files that will contain results in excel format
	nbCross    = int(sys.argv[4])  # cross validation identifier
	nbFold     = int(sys.argv[5])  # number of folds of the cross validation
	factor     = sys.argv[6]       # create factor ambiguous examples from each complete example    
	ambiguity  = sys.argv[7]       # level of ambiguity (percentage)
	nbErrors   = sys.argv[8]       # number of accepted errors on negative examples
	nbTries    = sys.argv[9]       # number of tries to enhance an hypothesis
                                    # having less than nbErrors
	compWeight = sys.argv[10]	 # weight of the only compatible examples
	doubleRep  = sys.argv[11]	 # double representation (double_representation) 	

	print("begin of the script ", script)

	suffixProba = "_p" + ambiguity
	fileExcelProba = fileExcel + suffixProba + ".csv"
	fileExcel = fileExcel + suffixProba
	directoryProba = fileEx    + suffixProba

	os.mkdir(directoryProba)
	os.chdir(directoryProba)	
					
	processes = []

							# creation des process crossValidation
	for cross in range(0, nbCross) :	
		processes.append(
			             mp.Process(
						         target = crossValidation, 
						         name = "lear_p" + ambiguity +
                                              "_n" + str(cross), 
						         args = (
							             fileEx, fileBkg, 
                                                fileExcel, fileExcelProba,
                                                cross, nbFold, factor, ambiguity,
                                                nbErrors, nbTries, compWeight,
									   doubleRep
							            )
                                     )
                          )   						

							# running the cross validations	  
	for p in processes : 
		p.start()      

							# waiting for the cross validations terminate
	for p in processes : 	
	 	p.join()       

	print("end of the script ", script)

# ====================================== END ===================================
