#!/usr/bin/python3.2
# (home + travail + maggi)


# ==============================================================================
#                       CREATE N * K CONCURRENT CROSS-VALIDATIONS USING NFOIL		
#                        (version sans parallelisme)
#
#	script     = sys.argv[0]       # nom du script
#	fileEx     = sys.argv[1]       # file containing intentionsal complete examples (.int)
#    fileBkg    = sys.argv[2]       # file containing background knowledge (.int_bkg)   
#	fileExcel  = sys.argv[3]       # prefix of files that will contain results in excel format
#	nbCross    = int(sys.argv[4])  # cross validation identifier
#	nbFold     = int(sys.argv[5])  # number of folds of the cross validation 
#	ambiguity  = sys.argv[6]       # level of ambiguity (percentage)

# =================================== IMPORTS ==================================

import sys
import os
import multiprocessing as mp
import subprocess
import lear_script_utilities as utilities
import shutil

# =============================== VARIABLES GLOBALES ===========================

learDirectory 		= os.getenv('LEAR_DIR')
nfoilDirectory 	= os.getenv('NFOIL_DIR')
LEAR 			= os.path.join(learDirectory, "topLear3.pl")
initialDirectory 	= os.getcwd()

# =================================== FONCTIONS ================================
				
					# make a nbFold cross validation
def crossValidation(fileEx, fileBkg, fileExcel, fileExcelProba,
       			cross, nbFold, ambiguity) :
			
					# we are in the proba directory

	processName = mp.current_process().name
	suffixCross = "_n" + str(cross)
	fileExcelCross = fileExcel + suffixCross

	currentDirectory = os.getcwd()
	baseName = os.path.basename(currentDirectory)
	baseCrossDirectory = baseName + suffixCross
	crossDirectory = os.path.join(currentDirectory, baseCrossDirectory)
	os.chdir(crossDirectory)	

					# we are in the cross directory
	pOut = open(processName + ".out", "w")
	pErr = open(processName + ".err", "w")
	fileExcel1 = os.path.join(crossDirectory, fileExcelCross + ".csv")

	for fold in range(1, nbFold+1) :	

		print("fold : ", fold),
		foldBase = baseCrossDirectory + "_fold" + str(fold)
							
					# go to the directory of the current fold
		foldDirectory = os.path.join(crossDirectory, foldBase) 
		os.chdir(foldDirectory)
		
					# copy intentional background file to the current
					# directory

		intFileBkg = os.path.join(initialDirectory, fileBkg + ".int_bkg")
		shutil.copy(intFileBkg, foldDirectory)

					# build the name of the ambiguity and nfoil files"
		file_ambiguity  = foldBase + "_train_ambiguous_close_world"	
		file_test  = foldBase + "_test"         
		nFoilData = file_ambiguity + "_nfoil"
		nFoilParameters = "nfoil_cross_validation_one_proba.cmd"

					# convert train and test examples to nfoil format
		fileName = os.path.join(learDirectory, "learToNfoilTrainTest.pl")
		args = [fileName, file_ambiguity, file_test, fileBkg, 
 			   nFoilData, nFoilParameters]
		subprocess.call(args, stdout = pOut, stderr = pErr)
		
					# run nfoil								
		nFoilProgram = os.path.join(nfoilDirectory, "nfoil")     
		fNfoilParameters = open(nFoilParameters, 'r')
		line = fNfoilParameters.readline()
		fNfoilParameters.close()

#		os.remove("nfoil_cross_validation_one_proba.cmd")
		args = line.split(' ')
		args.insert(0, nFoilProgram)

		resultsFileName = "nfoil.out"
		errorFileName = "nfoil.err"
		
		nFoilOut = open(resultsFileName, "w")
		nFoilErr = open(errorFileName, "w")
		
		subprocess.call(args, stdout = nFoilOut, stderr = nFoilErr)

		nFoilOut.close()
		nFoilErr.close()

					# get the results and write them
		(truePos, falseNeg, trueNeg, falsePos) = utilities.getNfoilResults(resultsFileName)
		
		fExcel = open(fileExcel1, "a")		
		fExcel.write(truePos + ";" + falseNeg + ";" + 
				     trueNeg + ";" + falsePos + "\n")
		fExcel.close()
		
					# return to the cross directory		
		os.chdir("..")	 
		
					# get the average of the results of the nbFold folds
	crossValidationResults = utilities.getCrossValidationResults(fileExcel1)

					# return to the proba directory
	os.chdir("..")		
		
					# save results
	utilities.addCrossValidationResults(fileExcelProba, crossValidationResults)

	fExcel.close()
	pOut.close()
	pErr.close()

# ===================================== MAIN ===================================

if __name__ == '__main__' :

	script     = sys.argv[0]       # nom du script
	fileEx     = sys.argv[1]       # file containing intentionsal complete examples (.int)
	fileBkg    = sys.argv[2]       # file containing background knowledge (.int_bkg)   
	fileExcel  = sys.argv[3]       # prefix of files that will contain results in excel format
	nbCross    = int(sys.argv[4])  # cross validation identifier
	nbFold     = int(sys.argv[5])  # number of folds of the cross validation 
	ambiguity  = sys.argv[6]       # level of ambiguity (percentage)

	print("begin of the script ", script)
	print("probabilite = ", ambiguity)
	print("current directory = " + os.getcwd())
	print()
	print()

	suffixProba = "_p" + ambiguity
	fileExcelProba = fileExcel + suffixProba + ".csv"
	fileExcel = fileExcel + suffixProba
	directoryProba = fileEx + suffixProba
	os.chdir(directoryProba)	

								# we are in the proba directory
	print("proba directory = " + directoryProba)

	processes = []

								# create the processes
	for cross in range(0, nbCross) :

		processes.append(
			             mp.Process(
						         target = crossValidation, 
						         name = "nfoil_p" + ambiguity + 
                                              "_n" + str(cross), 
						         args = (
							             fileEx, fileBkg, 
                                                fileExcel, fileExcelProba,
                                                cross, nbFold, ambiguity
							            )
                                     )
                          )   			

								# running the cross validations	  
	for p in processes : 
		p.start()      

								# waiting for the cross validations terminate
	for p in processes : 	
	 	p.join()       


	os.chdir("..")

	print("end of the script ", script)

# ====================================== END ===================================
