#!/usr/bin/python3.2
# (home + travail + magi)

# ==============================================================================
#                       CREATE N CONCURRENT K CROSS-VALIDATIONS USING TILDE						   
#
#	script     = sys.argv[0]       # nom du script
#	fileEx     = sys.argv[1]       # file containing train intentionsal complete examples (.int)
#	fileExcel  = sys.argv[2]       # prefix of files that will contain results in excel format
#	nbCross    = int(sys.argv[3])  # cross validation identifier
#	nbFold     = int(sys.argv[4])  # number of folds of the cross validation 
#	ambiguity  = sys.argv[5]       # level of ambiguity (percentage)

# =================================== IMPORTS ==================================

import sys
import os
import multiprocessing as mp
import subprocess
import lear_script_utilities as utilities

# =============================== VARIABLES GLOBALES ===========================

learDirectory = os.getenv('LEAR_DIR')
LEAR = os.path.join(learDirectory, "topLear3.pl")
lockFileProba = mp.Lock()

# =================================== FONCTIONS ================================

	# make a nbFolds cross validation
def crossValidation(fileEx, fileExcel, fileExcelProba,
       			cross, nbFold, ambiguity) :
	
						# we are in the proba directory
	suffixCross = "_n" + str(cross)
	fileExcelCross = fileExcel + suffixCross
	processName = mp.current_process().name
	
						# go to the cross directory
	currentDirectory = os.getcwd()
	baseName = os.path.basename(currentDirectory)
	baseCrossDirectory = baseName + suffixCross
	crossDirectory = os.path.join(currentDirectory, baseCrossDirectory)
	os.chdir(crossDirectory)	
	
						# we are in the cross directory
	pOut = open(processName + ".out", "w")
	pErr = open(processName + ".err", "w")
	fileExcel1 = os.path.join(crossDirectory, fileExcelCross + ".csv")
	          
						# for each fold	         						     		               	
	for fold in range(1, nbFold+1) :
			
		print("fold : ", fold),
		foldBase = baseCrossDirectory + "_fold" + str(fold)
							
						# go to the fold directory
		foldDirectory = os.path.join(crossDirectory, foldBase) 
		os.chdir(foldDirectory)

						# build the name of the ambiguity file
		file_ambiguity  = foldBase + "_train_ambiguous_close_world"	
		file_test       = foldBase + "_test"  
      
						# copy file containing the parameters
		os.system("cp ../../../" + fileEx + ".s "  + file_ambiguity + "_tilde.s")
						# copy file containing the background 	
		os.system("cp ../../../" + fileEx + ".bg " + file_ambiguity + "_tilde.bg") 
		
						# convert train and test examples to Tilde format
		print("learToTilde.pl")
		programme = os.path.join(learDirectory, "learToTilde.pl")
		args = [programme, file_ambiguity, file_test, file_ambiguity + "_tilde"]
		subprocess.call(args, stdout = pOut, stderr = pErr)
		print("fin learToTilde.pl")
		
						# create Tilde commands file
		fTildeInput = open('tilde_cross_validation_one_proba.cmd', 'w')
		fTildeInput.write('leave_one_out_from_list(tilde, [1])\n')
		fTildeInput.write('q\n')
		fTildeInput.close()

						# run Tilde 	
		print("run Tilde")							
		aceDirectory = os.getenv('ACE_DIR')
		programme = os.path.join(aceDirectory, "ace")     
		fTildeInput = open('tilde_cross_validation_one_proba.cmd', 'r')
		subprocess.call(programme, 
                        stdin = fTildeInput, stdout = pOut, stderr = pErr),
		fTildeInput.close()
		print("fin Tilde.")

		os.remove("tilde_cross_validation_one_proba.cmd")
		
						# get the results and write them
		(truePos, falseNeg, trueNeg, falsePos) = utilities.getTildeResults(file_ambiguity)		
		fExcel = open(fileExcel1, "a")		
		fExcel.write(truePos + ";" + falseNeg + ";" + 
				     trueNeg + ";" + falsePos + "\n")
		fExcel.close()
		
		os.chdir("..")		# return to the cross directory		
		
						# get the average of the results of the folds
	crossValidationResults = utilities.getCrossValidationResults(fileExcel1)

	os.chdir("..")			# return to the proba directory
		

						# save results for this cross validation
	lockFileProba.acquire()
	utilities.addCrossValidationResults(fileExcelProba, crossValidationResults)
	lockFileProba.release()

	fExcel.close()
	pOut.close()
	pErr.close()


# ===================================== MAIN ===================================

if __name__ == '__main__' :

	script     = sys.argv[0]       # nom du script
	fileEx     = sys.argv[1]       # file containing train intentionsal complete examples (.int)
	fileExcel  = sys.argv[2]       # prefix of files that will contain results in excel format
	nbCross    = int(sys.argv[3])  # cross validation identifier
	nbFold     = int(sys.argv[4])  # number of folds of the cross validation 
	ambiguity  = sys.argv[5]       # level of ambiguity (percentage)

	print("begin of the script ", script)
	print("current directory = " + os.getcwd())

	suffixProba 	= "_p" + ambiguity
	fileExcelProba = fileExcel + suffixProba + ".csv"
	fileExcel 	= fileExcel + suffixProba
	directoryProba = fileEx    + suffixProba

							# go to the proba directory
	os.chdir(directoryProba)	
							# creation des processes crossValidation					
	processes = []
							
	for cross in range(0, nbCross) :	
	
		processes.append(
			             mp.Process(
						               target = crossValidation, 
						               name = "tilde_p" + ambiguity + 
                                                    "_n" + str(cross), 
						               args = (
							                    fileEx, fileExcel, fileExcelProba,
                                                       cross, nbFold, ambiguity
							                   )
                                      )
                          )   						

							# running the cross validations	  
	for p in processes : 
		p.start()      

							# waiting for the cross validations terminate
	for p in processes : 	
	 	p.join()       

print("end of the script ", script)

# ====================================== END ===================================
